import sys
import random
import argparse
import functools
import logging

import os
import torch
import numpy as np

from .sessions import train, test
from .models import factory
from .io import get_dataset
import pathlib


class Main:
    arguments = {
        ('action', ): {'type': str, 'help': 'Name of the action to perform.'},
        ('dataset', ): {'type': str, 'help': 'Name of the dataset.'},
        ('model', ): {'type': str, 'help': 'Name of the model.'},

        # dataset path
        ('-dp', '--dataset-path'): {
            'type': str, 'default': os.path.join(pathlib.Path(__file__).parent.parent.resolve(),'datasets'), 'help': 'Path to the dataset.',
        },

        # checkpointing
        ('-load', '--load-name'): {
            'type': str, 'default': None,
            'help': 'Name of the saved model to restore.'
        },
        ('-save', '--save-name'): {
            'type': str, 'default': None,
            'help': 'Name of the saved model to save.'
        },

        # common training args
        ('-opt', '--optimizer'): {
            'type': str, 'default': 'adam', 'help': 'Pick an optimizer.',
        },
        ('-lr', '--learning-rate'): {
            'type': float, 'default': 0.01, 'help': 'Initial learning rate.',
        },
        ('-m', '--max-epochs'): {
            'type': int, 'default': 100,
            'help': 'Maximum number of epochs for training.',
        },
        ('-b', '--batch-size'): {
            'type': int, 'default': 1,
            'help': 'Batch size for training and evaluation.',
        },

        # debug control
        ('-d', '--debug'): {
            'action': 'store_true', 'help': 'Verbose debug',
        },
        ('-seed', '--seed'): {
            'type': int, 'default': 2, 'help': 'Number of steps for model optimisation',
        },

        # cpu gpu setup
        ('-w', '--num_workers'): {
            # multiprocessing fail with too many works
            'type': int, 'default': 0, 'help': 'Number of CPU workers.',
        },
        ('-n', '--num_devices'): {
            'type': int, 'default': 1, 'help': 'Number of GPU devices.',
        },
        ('-a', '--accelerator'): {
            'type': str, 'default': 'gpu', 'help': 'Accelerator style.',
        },
        ('-s', '--strategy'): {
            'type': str, 'default': 'ddp', 'help': 'strategy style.',
        }
    }
    def __init__(self):
        super().__init__()
        a = self.parse()
        if a.debug:
            sys.excepthook = self._excepthook
        # seeding
        random.seed(a.seed)
        torch.manual_seed(a.seed)
        np.random.seed(a.seed)
        self.a = a

    def parse(self):
        p = argparse.ArgumentParser(description='Genome Graph.')
        for k, v in self.arguments.items():
            p.add_argument(*k, **v)
        return p.parse_args()

    def _excepthook(self, etype, evalue, etb):
        from IPython.core import ultratb
        ultratb.FormattedTB()(etype, evalue, etb)
        for exc in [KeyboardInterrupt, FileNotFoundError]:
            if issubclass(etype, exc):
                sys.exit(-1)
        import ipdb
        ipdb.post_mortem(etb)

    def run(self):
        try:
            action = getattr(self, f'cli_{self.a.action.replace("-", "_")}')
        except AttributeError:
            callables = [n[4:] for n in dir(self) if n.startswith('cli_')]
            logging.error(
                f'Unkown action {self.a.action!r}, '
                f'accepts: {", ".join(callables)}.')
        return action()

    def setup_model_and_data(self, a):
        # get dataset
        print(f"datasets path: {a.dataset_path}")
        train_loader, val_loader, test_loader, dataset_info = get_dataset(
            name=a.dataset, batch_size=a.batch_size, workers=a.num_workers,
            datasets_path=a.dataset_path)
        # get model
        model_cls = factory[a.model]
        model = model_cls(info=dataset_info)
        return model, train_loader, val_loader, test_loader, dataset_info

    def cli_train(self):
        a = self.a
        if not a.save_name:
            logging.error('--save-name not specified.')

        model, train_loader, val_loader, test_loader, dataset_info = self.setup_model_and_data(a)

        plt_trainer_args = {
            'max_epochs': a.max_epochs, 'devices': a.num_devices,
            'accelerator': a.accelerator, 'strategy': a.strategy,
            'fast_dev_run': a.debug,}

        train_params = {
            'model': model,
            'train_loader': train_loader,
            'val_loader': val_loader,
            'optimizer': a.optimizer,
            'learning_rate': a.learning_rate,
            "plt_trainer_args": plt_trainer_args,
            "save_path": a.save_name,
            "dataset_info": dataset_info,
            "seed": a.seed,
        }
        train(**train_params)

    def cli_test(self):
        a = self.a

        model, train_loader, val_loader, test_loader, dataset_info = self.setup_model_and_data(a)
        plt_trainer_args = {
            'devices': a.num_devices,
            'accelerator': a.accelerator, 'strategy': a.strategy,}
        test_params = {
            'model': model,
            'test_loader': test_loader,
            'plt_trainer_args': plt_trainer_args,
            'load_path': a.load_name,
            'dataset_info': dataset_info,
        }
        test(**test_params)
    cli_eval = cli_test

# hybrid_graph/cli.py
def main():
    Main().run()



import torch


class RandomNodeSampler(torch.utils.data.DataLoader):
    # Graphs sampled using this method can be very sparse.
    def __init__(self, data, num_samples, sample_size, mask=None, **kwargs):
        # Remove for PyTorch Lightning:
        # TODO: Copied from torch_geometric.loader.graph_saint.GraphSAINTSampler. Not sure if it is needed
        kwargs.pop('dataset', None)
        kwargs.pop('collate_fn', None)

        self.data = data
        self.num_samples = num_samples
        self.sample_size = sample_size
        self.mask = mask

        super().__init__(self, collate_fn=self._collate, **kwargs)

    def __len__(self):
        return self.num_samples

    def __getitem__(self, item):
        # Apply the mask if provided
        if self.mask is not None:
            masked_nodes = torch.arange(self.data.num_nodes)[self.mask]
        else:
            masked_nodes = torch.arange(self.data.num_nodes)

        # Sample random nodes
        perm = torch.randperm(masked_nodes.size(0))
        shuffled_nodes = masked_nodes[perm]
        sampled_nodes = shuffled_nodes[:self.sample_size]

        # Get the subgraph induced by the sampled nodes
        x_sampled = self.data.x[sampled_nodes]
        y_sampled = self.data.y[sampled_nodes]

        # Get edges and hyperedges between the sampled nodes
        edge_mask = torch.isin(self.data.edge_index[0], sampled_nodes) & \
                    torch.isin(self.data.edge_index[1], sampled_nodes)
        edge_index_sampled = self.data.edge_index[:, edge_mask]
        hyperedge_mask = torch.isin(self.data.hyperedge_index[0], sampled_nodes)
        hyperedge_index_sampled = self.data.hyperedge_index[:, hyperedge_mask]
        _, inverse_indices = torch.unique(hyperedge_index_sampled[1], return_inverse=True)
        hyperedge_index_sampled[1] = inverse_indices

        # Remap node indices to the range [0, sample_size - 1]
        node_map = {node: i for i, node in enumerate(sampled_nodes.tolist())}
        edge_index_sampled.apply_(lambda node: node_map[node])
        hyperedge_index_sampled[0].apply_(lambda node: node_map[node])

        return x_sampled, y_sampled, edge_index_sampled, hyperedge_index_sampled

    def _collate(self, data_list):
        assert len(data_list) == 1
        x_sampled, y_sampled, edge_index_sampled, hyperedge_index_sampled = data_list[0]

        data = self.data.__class__()
        data.num_nodes = x_sampled.size(0)
        data.x = x_sampled
        data.y = y_sampled
        data.edge_index = edge_index_sampled
        data.hyperedge_index = hyperedge_index_sampled
        data.num_hyperedges = hyperedge_index_sampled[1, -1] + 1

        return data


class RandomHyperedgeSampler(torch.utils.data.DataLoader):
    # Graphs sampled using this method can be dense.
    # WARNING: THIS METHOD CAN GET VERY SLOW FOR LARGE SAMPLES.
    def __init__(self, data, num_samples, min_sample_size, mask=None, **kwargs):
        # Remove for PyTorch Lightning:
        # TODO: Copied from torch_geometric.loader.graph_saint.GraphSAINTSampler. Not sure if it is needed
        kwargs.pop('dataset', None)
        kwargs.pop('collate_fn', None)

        self.data = data
        self.num_samples = num_samples
        self.min_sample_size = min_sample_size
        self.mask = mask

        super().__init__(self, collate_fn=self._collate, **kwargs)

    def __len__(self):
        return self.num_samples

    def __getitem__(self, item):
        # Sample hyperedges until they contain at least min_sample_size nodes, and apply the mask if provided
        perm = torch.randperm(self.data.num_hyperedges)
        sampled_nodes_mask = torch.zeros(self.data.num_nodes)
        i = 0
        while int(torch.sum(sampled_nodes_mask)) < self.min_sample_size:
            sample = perm[i]
            sampled_nodes_mask[self.data.hyperedge_index[0, self.data.hyperedge_index[1, :] == sample]] = 1
            if self.mask is not None:
                sampled_nodes_mask = sampled_nodes_mask & self.mask
            i += 1
        sampled_nodes = torch.arange(self.data.num_nodes)[sampled_nodes_mask.type(torch.bool)]

        # Get the subgraph induced by the sampled nodes
        x_sampled = self.data.x[sampled_nodes]
        y_sampled = self.data.y[sampled_nodes]

        # Get edges and hyperedges between the sampled nodes
        edge_mask = torch.isin(self.data.edge_index[0], sampled_nodes) & \
                    torch.isin(self.data.edge_index[1], sampled_nodes)
        edge_index_sampled = self.data.edge_index[:, edge_mask]
        hyperedge_mask = torch.isin(self.data.hyperedge_index[0], sampled_nodes)
        hyperedge_index_sampled = self.data.hyperedge_index[:, hyperedge_mask]
        _, inverse_indices = torch.unique(hyperedge_index_sampled[1], return_inverse=True)
        hyperedge_index_sampled[1] = inverse_indices

        # Remap node indices to the range [0, sample_size - 1]
        node_map = {node: i for i, node in enumerate(sampled_nodes.tolist())}
        edge_index_sampled.apply_(lambda node: node_map[node])
        hyperedge_index_sampled[0].apply_(lambda node: node_map[node])

        return x_sampled, y_sampled, edge_index_sampled, hyperedge_index_sampled

    def _collate(self, data_list):
        assert len(data_list) == 1
        x_sampled, y_sampled, edge_index_sampled, hyperedge_index_sampled = data_list[0]

        data = self.data.__class__()
        data.num_nodes = x_sampled.size(0)
        data.x = x_sampled
        data.y = y_sampled
        data.edge_index = edge_index_sampled
        data.hyperedge_index = hyperedge_index_sampled
        data.num_hyperedges = hyperedge_index_sampled[1, -1] + 1

        return data



from .gnn.baseline import GCNNet, SAGENet
from .gnn.gat import GATNet, GATV2Net
from .gnn.hyper import HyperGCN, HyperGAT
from .gnn.linearprobe import LPGCNHyperGCN,LPGGATGCN, LPGATHyperGCN, LPHYPERHYPER, LPGCNGCN, LPGGATGAT
from .ensemble.average_prediction import Average_Ensemble

factory = {
    'gcn': GCNNet,
    'sage': SAGENet,
    'gat': GATNet,
    'gatv2': GATV2Net,
    'hyper-gcn': HyperGCN,
    'hyper-gat': HyperGAT,
    'ensemble': Average_Ensemble,
    'lp-gcn-hyper-gcn': LPGCNHyperGCN,
    'lp-gat-hyper-gcn': LPGATHyperGCN,
    'lp-gat-gcn': LPGGATGCN,
    'lp-gcn-gcn': LPGCNGCN,
    'lp-gat-gat': LPGGATGAT,
    'lp-hyper-hyper': LPHYPERHYPER,
}



import os
from typing import Callable, Optional

import numpy as np
import torch
from torch_geometric.data import Data, InMemoryDataset, download_url


class Wikipedia(InMemoryDataset):
    r"""The Wikipedia networks introduced in the
    `"Multi-scale Attributed Node Embedding"<https://arxiv.org/abs/1909.13021>`_ paper.
    Nodes represent web pages and edges represent mutual hyperlinks between them. Hyperedges are mutually linked
    page groups that contain at least 3 pages (i.e., maximal cliques with sizes of at least 3).
    Node features represent several informative nouns in the Wikipedia pages.
    The task is to predict the average monthly traffic of the web page.

    Args:
        root (str): Root directory where the dataset should be saved.
        name (str): The name of the dataset (:obj:`"chameleon"`, :obj:`"crocodile"`, :obj:`"squirrel"`).
        use_musae_node_embeddings (bool): If set to :obj:`True`, will load the pre-processed node embeddings
            as introduced in the `"Multi-scale Attributed Node Embedding"<https://arxiv.org/abs/1909.13021>`_ paper.
            If set to :obj:`False`, will load the raw node features extracted based on informative nouns appeared in
            the text of the Wikipedia articles. (default: :obj:`True`)
        transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object
            and returns a transformed version. The data object will be transformed before every access.
            (default: :obj:`None`)
        pre_transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data`
            object and returns a transformed version. The data object will be transformed before being saved to disk.
            (default: :obj:`None`)

    **STATS:**

    .. list-table::
        :widths: 15 15 15 15
        :header-rows: 1

        * - Name
          - #nodes
          - #edges
          - #hyperedges
          - #features
        * - chameleon
          - 2,277
          - 62,742
          - 14,650
          - 128 or 3,132
        * - crocodile
          - 11,631
          - 341,546
          - 121,431
          - 128 or 13,183
        * - squirrel
          - 5,201
          - 396,706
          - 220,678
          - 128 or 3,148
    """
    url = "https://drive.google.com/uc?export=download&confirm=no_antivirus&id={}"
    file_id = {
        'chameleon': '1qLV8dMIhw-pe4ym16moOS9rl5rLD0FQr',
        'crocodile': '1D9c4D-V4iIxjklfaACGilrw_H0-rGVU5',
        'squirrel': '1-SBN-KSgSTpVJibG65slJDsBCBzXBj18',
    }

    NUM_HYPEREDGES = {
        'chameleon': 14650,
        'crocodile': 121431,
        'squirrel': 220678,
    }

    def __init__(self, root: str, name: str, use_musae_node_embeddings: bool = True,
                 transform: Optional[Callable] = None, pre_transform: Optional[Callable] = None):
        self.name = name.lower()
        self.use_musae_node_embeddings = use_musae_node_embeddings
        assert self.name in ['chameleon', 'crocodile', 'squirrel']
        super().__init__(root, transform, pre_transform)
        self.data, self.slices = torch.load(self.processed_paths[0])

    @property
    def raw_dir(self) -> str:
        return os.path.join(self.root, self.name, 'raw')

    @property
    def processed_dir(self) -> str:
        if self.use_musae_node_embeddings:
            return os.path.join(self.root, self.name, 'processed', 'musae_node_embeddings')
        else:
            return os.path.join(self.root, self.name, 'processed', 'raw_node_features')

    @property
    def raw_file_names(self) -> str:
        return f'wikipedia_{self.name}.npz'

    @property
    def processed_file_names(self) -> str:
        return 'data.pt'

    def download(self):
        download_url(self.url.format(self.file_id[self.name]), self.raw_dir, filename=self.raw_file_names)

    def process(self):
        data = np.load(self.raw_paths[0], 'r', allow_pickle=True)
        if self.use_musae_node_embeddings:
            x = torch.from_numpy(data['features']).to(torch.float)
        else:
            x = torch.from_numpy(data['raw_features']).to(torch.float)
        y = torch.from_numpy(data['target']).to(torch.float)
        edge_index = torch.from_numpy(data['edges']).to(torch.long)
        hyperedge_index = torch.from_numpy(data['hyperedges']).to(torch.long)

        data = Data(x=x, y=y, edge_index=edge_index, hyperedge_index=hyperedge_index,
                    num_hyperedges=self.NUM_HYPEREDGES[self.name])

        if self.pre_transform is not None:
            data = self.pre_transform(data)

        torch.save(self.collate([data]), self.processed_paths[0])



import yaml
import torch
import torch.nn.functional as F
import torch.nn as nn
from torch_geometric.nn import GCNConv, SAGEConv

"""
for baseline method:
# self.hyperGCN
# x2 = hyperGCN(x,hyperedge_index)
# x,x2
# potential 2 impelmentations
"""
class GCNNet(torch.nn.Module):
    def __init__(
            self, info, *args, **kwargs):
        super().__init__()
        dim = 32
        self.conv1 = GCNConv(info["num_node_features"], dim)
        self.is_regression = info["is_regression"]
        self.is_edge_pred = info["is_edge_pred"]
        if info["is_regression"]:
            self.conv2 = GCNConv(dim, dim)
            self.head = nn.Linear(dim, 1)
        elif info["is_edge_pred"]:
            self.conv2 = GCNConv(dim, dim)
        else:
            self.conv2 = GCNConv(dim, info["num_classes"])

    def forward(self, data, *args, **kargs):
        x, edge_index = data.x, data.edge_index
        if self.is_edge_pred:
            edge_index = args[0][0] # the message passing edge index
        x = F.relu(self.conv1(x, edge_index))
        x = F.dropout(x, training=self.training)
        x = self.conv2(x, edge_index)


        if self.is_regression:
            x = self.head(x).squeeze()
        elif self.is_edge_pred:
            # args[1] is the edge_label_index
            edge_label_index = args[0][1]
            x = x[edge_label_index[0]]*x[edge_label_index[1]]
            x = torch.sum(x,-1)
        else:
            x = F.log_softmax(x, dim=1)
        return x


class SAGENet(torch.nn.Module):
    def __init__(
            self, info, mixture_cls=None, load_config=None, *args, **kwargs):
        super(SAGENet, self).__init__()
        dim = 32
        self.is_regression = info["is_regression"]
        self.is_edge_pred = info["is_edge_pred"]
        if self.is_regression:
            self.conv2 = SAGEConv(dim, dim, normalize=False)
            self.head = nn.Linear(dim, 1)
        elif info["is_edge_pred"]:
            self.conv2 = SAGEConv(dim, dim)
        else:
            self.conv2 = SAGEConv(dim, info["num_classes"], normalize=False)

        self.conv1 = SAGEConv(
            info["num_node_features"], dim, normalize=False)

    def forward(self, data, *args, **kargs):
        x, edge_index = data.x, data.edge_index
        if self.is_edge_pred:
            edge_index = args[0][0] # the message passing edge index
        x = F.relu(
            self.conv1(x, edge_index))
        x = F.dropout(x, p=0.5, training=self.training)
        x = self.conv2(x, edge_index)
        if self.is_regression:
            x = self.head(x).squeeze()
        elif self.is_edge_pred:
            # args[1] is the edge_label_index
            edge_label_index = args[0][1]
            x = x[edge_label_index[0]]*x[edge_label_index[1]]
            x = torch.sum(x,-1)
        else:
            x = F.log_softmax(x, dim=1)
        return x



import os
import sys
import math
import signal
import functools
import numpy as np
import pathlib
import torch
import torch_geometric
import torch_geometric.utils as utils
import logging

from .utils import device
import yaml
import hg.datasets as datasets
torch.multiprocessing.set_sharing_strategy('file_system')


class DataLoader(torch_geometric.loader.DataLoader):
    pin_memory = True

    def __init__(
            self, dataset, masks, batch_size,
            workers, single_graph=False, shuffle=False, onehot=False, sampler=None, batch_sampler=None):
        super().__init__(
            dataset, batch_size,
            pin_memory=self.pin_memory,
            num_workers=workers, worker_init_fn=self.worker_init,
            shuffle=shuffle, sampler=sampler, batch_sampler=None)
        self.single_graph = single_graph
        self.masks = masks
        self.onehot = onehot
        self.workers = workers
        # self.sampler = sampler
        # TODO sampler is potentially needed
        # self.sampler = None
        # self.batch_sampler = None

    @staticmethod
    def worker_init(x):
        signal.signal(signal.SIGINT, signal.SIG_IGN)

    # TODO implement hyper-graph partition
    def __iter__(self):
        if self.single_graph:
            for i, item in enumerate(super().__iter__()):
                yield item
        else:
            for i, item in enumerate(self.sampler.__iter__()):
                yield item


def get_dataset(name, datasets_path=os.path.join(pathlib.Path(__file__).parent.parent.parent.resolve(),'datasets'),
                original_mask=False, split=0.6, batch_size=6000, workers=2, num_steps=5):
    # if datasets_path not in sys.path:
    #     sys.path.append(datasets_path)
    # import datasets

    with open(os.path.join(datasets_path, 'dataset_info.yaml')) as f:
        DATASET_INFO = yaml.safe_load(f)

    # fix random seeds
    np.random.seed(1)
    torch.manual_seed(1)

    info = dict(DATASET_INFO[name])
    dataset_info = info.pop('info', {})

    single_graph = info.pop('single_graph', False)
    onehot = info.pop('onehot', False)

    cls = getattr(datasets, info.pop('type'))
    print(info)
    dataset = cls(**info)
    kwargs = {
        'batch_size': 1,
        'workers': workers,
        'single_graph': single_graph,
    }


    original_mask = dataset_info.pop('original_mask')
    Loader = functools.partial(DataLoader, **kwargs)
    if dataset_info['is_edge_pred']:
        dataset = datasets.create_edge_label(dataset)
    dataset, masks = datasets.mask_split(dataset, original_mask)
    # take one sample mask out
    train_mask, eval_mask, test_mask = masks[0]
    dataset = dataset[0]
    dataset.train_mask = train_mask
    dataset.val_mask = eval_mask
    dataset.test_mask = test_mask
    print(dataset)
    # dataloader requires a list of dataset
    dataset = [dataset]
    # logging.info(
    print(
        f"Search with a partition of {train_mask.sum()} train data, "
        f"{eval_mask.sum()} val data and {test_mask.sum()} test data.")
    # for single graph the masks is of no use
    print(dataset_info)
    if single_graph:
        return Loader(dataset, masks), Loader(dataset, masks), Loader(dataset, masks), dataset_info
    else:
        kwargs = {
            "batch_size": batch_size,
            "num_steps": num_steps,
            "num_workers": workers,
        }
        Sampler = functools.partial(datasets.HypergraphSAINTNodeSampler, **kwargs)
        return (
            Loader(dataset, masks, sampler=Sampler(dataset[0])),
            Loader(dataset, masks, sampler=Sampler(dataset[0])),
            Loader(dataset, masks, sampler=Sampler(dataset[0])),
            dataset_info,
        )



from torch_geometric.datasets import Planetoid
from ..hg_formatter.HybridGraphFormatter import GraphFormatter
class Benchmark:
    """Try to directly construct the hypergraph from a original embedding"""
    def __new__(self,root,name,k=None,percentile=80):
        assert name in ["Cora", "CiteSeer", "PubMed"], "Dataset not supported"
        dataset = Planetoid(root, name)
        data = GraphFormatter(k,percentile)(dataset[0])
        return [data]



from pytorch_lightning.callbacks import ModelCheckpoint
import pytorch_lightning as pl
from .plt_wrapper import ModelWrapper


def train(
        model,
        train_loader, val_loader,
        optimizer,
        learning_rate,
        plt_trainer_args,
        save_path,
        dataset_info,
        seed
        ):
    pl.seed_everything(seed, workers=True)
    checkpoint_callback = ModelCheckpoint(
        save_top_k=1,
        monitor="val_acc",
        mode="max",
        filename="best",
        dirpath=save_path,
        save_last=True,
    )
    plt_trainer_args['callbacks'] = [checkpoint_callback]
    if save_path is not None:
        plt_trainer_args['default_root_dir'] = f'./lightning_logs/{save_path.split("checkpoints/")[1]}'
    plt_model = ModelWrapper(
        model,
        dataset_info,
        learning_rate=learning_rate,
        epochs=plt_trainer_args['max_epochs'],
        optimizer=optimizer)
    trainer = pl.Trainer(**plt_trainer_args)
    trainer.fit(plt_model, train_loader, val_loader)



from .grand import Grand
from .musae import GitHub, Facebook, Twitch, Wikipedia
from .amazon import Amazon
from .standard import Benchmark, GraphStats
from .hg_spliter import mask_split, random_node_split, create_edge_label
from .hg_formatter import GraphFormatter
from .hg_samplers import (
    HypergraphSAINTSampler,
    HypergraphSAINTNodeSampler,
    HypergraphSAINTEdgeSampler,
    HypergraphSAINTRandomWalkSampler,
    RandomNodeSampler,
    RandomHyperedgeSampler,
)
__all__ = ['Grand', 'GitHub', 'Facebook', 'Twitch', 'Wikipedia', 'Amazon',
           'GraphFormatter', 'Benchmark', 'GraphStats',
           'mask_split', 'random_node_split', 'create_edge_label',
           'HypergraphSAINTSampler', 'HypergraphSAINTNodeSampler',
           'HypergraphSAINTEdgeSampler', 'HypergraphSAINTRandomWalkSampler',
           "RandomNodeSampler", "RandomHyperedgeSampler"]



from sklearn.manifold import TSNE
import matplotlib.pyplot as plt
import seaborn as sns
import torch
import networkx as nx
from torch_geometric.utils import to_networkx
from torch_geometric.data import Data
class EmbeddingVisualizer:
    def __init__(self, data, title='Embedding Visualizer'):
        self.embedding = data.x
        self.labels = data.y
        self.title = title
        self.tsne = TSNE(n_components=2, random_state=0)
        self.tsne_obj = self.tsne.fit_transform(self.embedding)
    def plot(self):
        plt.figure(figsize=(10, 8))
        sns.scatterplot(
            x=self.tsne_obj[:, 0],
            y=self.tsne_obj[:, 1],
            hue=self.labels,
            palette='Spectral',
            legend='full',
            s=60
        )
        plt.title(self.title)
        plt.savefig(self.title + ".png")
class HyperEdgeSizeHist:
    #TODO: need to be fixed
    def __init__(self, data, title='HyperEdge Size Histogram'):
        self.title = title
        self.counts = data.hyperedge_index[1].bincount()
    def plot(self):
        plt.figure(figsize=(10, 8))
        plt.hist(self.counts, log=True, alpha=0.5)
        plt.xlabel('HyperEdge Size')
        plt.ylabel('Count')
        plt.yscale('log')
        plt.title(self.title)
        plt.savefig(self.title + ".png")

class GraphStats:
    def __init__(self, data: Data, title='Graph Statistics'):
        self.data = data
        self.title = title
        self.graph = to_networkx(data, to_undirected=True) # Convert to a networkx graph for some computations
    def plot(self):
        #EmbeddingVisualizer(self.data).plot()
        HyperEdgeSizeHist(self.data).plot()
    def get_all_stats(self):
        # HyperEdgeSizeHist(self.data,self.title).plot()
        return {
            'title': self.title,
            'num_nodes': self.num_nodes(),
            'num_edges': self.num_edges(),
            'average_degree': self.average_degree(),
            'clustering_coefficient': self.clustering_coefficient(),
            'density': self.density(),
            # 'diameter': self.diameter(),
            'num_hyperedges': self.data.num_hyperedges,
            'average_size_of_hyperedge': self.average_size_of_hyperedge()
        }
    def average_size_of_hyperedge(self):
        return torch.sum(torch.bincount(self.data.hyperedge_index[1])) / self.data.num_hyperedges
    def num_nodes(self):
        return self.data.num_nodes

    def num_edges(self):
        return self.data.num_edges

    def average_degree(self):
        return 2 * self.num_edges() / self.num_nodes()

    def clustering_coefficient(self):
        # compute clustering coefficient with networkx
        return nx.average_clustering(self.graph)
    def density(self):
        # calculate density of graph
        n = self.num_nodes()
        m = self.num_edges()
        return 2*m / (n*(n-1))

    def diameter(self):
        if nx.is_connected(self.graph):
            diameter = nx.diameter(self.graph)
        else:
            diameter = None
        return diameter



import os
from typing import Callable, Optional, List

import numpy as np
import torch
from torch_geometric.data import Data, InMemoryDataset, download_url


class GitHub(InMemoryDataset):
    r"""The GitHub Web and ML Developers dataset introduced in the
    `"Multi-scale Attributed Node Embedding"<https://arxiv.org/abs/1909.13021>`_ paper.
    Nodes represent developers on GitHub and edges are mutual follower relationships. Hyperedges are mutually following
    developer groups that contain at least 3 developers (i.e., maximal cliques with sizes of at least 3).
    It contains 37,300 nodes, 578,006 edges, 223,672 hyperedges, 128 (if the MUSAE preprocessed node embeddings
    are used) or 4,005 (if the raw node features are used) node features, and 4 classes.

    Args:
        root (str): Root directory where the dataset should be saved.
        use_musae_node_embeddings (bool): If set to :obj:`True`, will load the pre-processed node embeddings
            as introduced in the `"Multi-scale Attributed Node Embedding"<https://arxiv.org/abs/1909.13021>`_ paper.
            If set to :obj:`False`, will load the raw node features extracted based on the location,
            repositories starred, employer and e-mail address. (default: :obj:`True`)
        transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object
            and returns a transformed version. The data object will be transformed before every access.
            (default: :obj:`None`)
        pre_transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data`
            object and returns a transformed version. The data object will be transformed before being saved to disk.
            (default: :obj:`None`)

    **STATS:**

    .. list-table::
        :widths: 15 15 15 15
        :header-rows: 1

        * - #nodes
          - #edges
          - #hyperedges
          - #features
          - #classes
        * - 37,700
          - 578,006
          - 223,672
          - 128 or 4,005
          - 4
    """

    url_data = 'https://graphmining.ai/datasets/ptg/github.npz'
    url_preprocessed = 'https://drive.google.com/uc?export=download&confirm=no_antivirus&id=1XfPUyxriHBI0s6UQeQ34BDnqdaQvKR7l'

    NUM_HYPEREDGES = 223672

    def __init__(self, root: str, use_musae_node_embeddings: bool = True,
                 transform: Optional[Callable] = None, pre_transform: Optional[Callable] = None):
        self.use_musae_node_embeddings = use_musae_node_embeddings
        super().__init__(root, transform, pre_transform)
        self.data, self.slices = torch.load(self.processed_paths[0])

    @property
    def processed_dir(self) -> str:
        if self.use_musae_node_embeddings:
            return os.path.join(self.root, 'processed', 'musae_node_embeddings')
        else:
            return os.path.join(self.root, 'processed', 'raw_node_features')

    @property
    def raw_file_names(self) -> List[str]:
        return ['github.npz', 'github_preprocessed.npz']

    @property
    def processed_file_names(self) -> str:
        return 'data.pt'

    def download(self):
        download_url(self.url_data, self.raw_dir)
        download_url(self.url_preprocessed, self.raw_dir, filename=self.raw_file_names[1])

    def process(self):
        data = np.load(self.raw_paths[0], 'r', allow_pickle=True)
        preprocessed = np.load(self.raw_paths[1], 'r', allow_pickle=True)
        if self.use_musae_node_embeddings:
            x = torch.from_numpy(data['features']).to(torch.float)
        else:
            x = torch.from_numpy(preprocessed['raw_features']).to(torch.float)
        y = torch.from_numpy(data['target']).to(torch.long)
        edge_index = torch.from_numpy(data['edges']).to(torch.long)
        edge_index = edge_index.t().contiguous()
        hyperedge_index = torch.from_numpy(preprocessed['hyperedges']).to(torch.long)
        # Create hyperedge_attr
        data = Data(x=x, y=y, edge_index=edge_index, hyperedge_index=hyperedge_index,
                    num_hyperedges=self.NUM_HYPEREDGES)

        if self.pre_transform is not None:
            data = self.pre_transform(data)

        torch.save(self.collate([data]), self.processed_paths[0])



import os
from typing import Callable, Optional

import numpy as np
import torch
from torch_geometric.data import Data, InMemoryDataset, download_url


class Twitch(InMemoryDataset):
    r"""The Twitch Gamer networks introduced in the
    `"Multi-scale Attributed Node Embedding"<https://arxiv.org/abs/1909.13021>`_ paper.
    Nodes represent gamers on Twitch and edges are followerships between them. Hyperedges are mutually following
    user groups that contain at least 3 gamers (i.e., maximal cliques with sizes of at least 3).
    The task is to predict whether a user streams mature content.
    Each dataset contains 128 (if the MUSAE preprocessed node embeddings are used) or 4,005 (if the raw node features
    are used) node features, and 2 classes.

    Args:
        root (str): Root directory where the dataset should be saved.
        name (str): The name of the dataset
            (:obj:`"DE"`, :obj:`"EN"`, :obj:`"ES"`, :obj:`"FR"`, :obj:`"PT"`, :obj:`"RU"`).
        use_musae_node_embeddings (bool): If set to :obj:`True`, will load the pre-processed node embeddings
            as introduced in the `"Multi-scale Attributed Node Embedding"<https://arxiv.org/abs/1909.13021>`_ paper.
            If set to :obj:`False`, will load the raw node features extracted based on the games played and liked,
            location and streaming habits. (default: :obj:`True`)
        transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object
            and returns a transformed version. The data object will be transformed before every access.
            (default: :obj:`None`)
        pre_transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data`
            object and returns a transformed version. The data object will be transformed before being saved to disk.
            (default: :obj:`None`)

    **STATS:**

    .. list-table::
        :widths: 15 15 15 15 15
        :header-rows: 1

        * - Name
          - #nodes
          - #edges
          - #hyperedges
          - #features
          - #classes
        * - DE
          - 9,498
          - 306,276
          - 297,315
          - 128 or 3,170
          - 2
        * - EN
          - 7,126
          - 70,648
          - 13,248
          - 128 or 3,170
          - 2
        * - ES
          - 4,648
          - 118,764
          - 77,135
          - 128 or 3,170
          - 2
        * - FR
          - 6,549
          - 225,332
          - 172,653
          - 128 or 3,170
          - 2
        * - PT
          - 1,912
          - 62,598
          - 74,830
          - 128 or 3,170
          - 2
        * - RU
          - 4,385
          - 74,608
          - 25,673
          - 128 or 3,170
          - 2
    """

    url = "https://drive.google.com/uc?export=download&confirm=no_antivirus&id={}"
    file_id = {
        'DE': '148A6DhddSz2qeWnlHwyTBJl8eLkhoPzv',
        'EN': '1vb8KrwvZvBHdT_dll6oe62HNk0EQfSOb',
        'ES': '1u-AXJ6En7ctBhPbwbW1dcCf946MmJWlS',
        'FR': '1WLpZmpQPhFIrnsHQ-4-8jxkYnorUIYS8',
        'PT': '13gVESn3AGbyoF8aveWrJeAPqUGwTG09o',
        'RU': '137tr2fJt-lKaQ_PcMSCl3gh33Gateke_'
    }

    NUM_HYPEREDGES = {
        'DE': 297315,
        'EN': 13248,
        'ES': 77135,
        'FR': 172653,
        'PT': 74830,
        'RU': 25673
    }

    def __init__(self, root: str, name: str, use_musae_node_embeddings: bool = True,
                 transform: Optional[Callable] = None, pre_transform: Optional[Callable] = None):
        self.name = name.upper()
        self.use_musae_node_embeddings = use_musae_node_embeddings
        assert self.name in ['DE', 'EN', 'ES', 'FR', 'PT', 'RU']
        super().__init__(root, transform, pre_transform)
        self.data, self.slices = torch.load(self.processed_paths[0])

    @property
    def raw_dir(self) -> str:
        return os.path.join(self.root, self.name, 'raw')

    @property
    def processed_dir(self) -> str:
        if self.use_musae_node_embeddings:
            return os.path.join(self.root, self.name, 'processed', 'musae_node_embeddings')
        else:
            return os.path.join(self.root, self.name, 'processed', 'raw_node_features')

    @property
    def raw_file_names(self) -> str:
        return f'twitch_{self.name}.npz'

    @property
    def processed_file_names(self) -> str:
        return 'data.pt'

    def download(self):
        download_url(self.url.format(self.file_id[self.name]), self.raw_dir, filename=self.raw_file_names)

    def process(self):
        data = np.load(self.raw_paths[0], 'r', allow_pickle=True)
        if self.use_musae_node_embeddings:
            x = torch.from_numpy(data['features']).to(torch.float)
        else:
            x = torch.from_numpy(data['raw_features']).to(torch.float)
        y = torch.from_numpy(data['target']).to(torch.long)
        edge_index = torch.from_numpy(data['edges']).to(torch.long)
        hyperedge_index = torch.from_numpy(data['hyperedges']).to(torch.long)

        data = Data(x=x, y=y, edge_index=edge_index, hyperedge_index=hyperedge_index,
                    num_hyperedges=self.NUM_HYPEREDGES[self.name])

        if self.pre_transform is not None:
            data = self.pre_transform(data)

        torch.save(self.collate([data]), self.processed_paths[0])



import os.path as osp
from typing import List, Callable, Optional
import torch
from torch_geometric.data import InMemoryDataset, download_url, Data


class Amazon(InMemoryDataset):
    r"""The Amazon dataset
    Args:
        root (str): Root directory where the dataset should be saved.
        num_graphs (int): The number of graphs to be loaded.
     **STATS:**

    .. list-table::
         :widths: 15 15 15 15 15
        :header-rows: 1

        * - Name
          - #nodes
          - #edges
          - #hyperedges
          - #features
          - #classes
        * - Photos
          - 6,777
          - 45,306
          - 6,777
          - 1000
          - 10
        * - Computers
          - 10,226
          - 55,324
          - 10,226
          - 1000
          - 10
        """
    url = "https://drive.google.com/uc?export=download&confirm=no_antivirus&id={}"

    file_id = {
        'Computers': '1wGIKse1qpeldFZyRTdJGlItMCOpGdVG2',
        'Photos': '1twcjbZjZDhc6mzD8yBtLKc_JTlRojmMz',
    }

    def __init__(self,  root: str, name: str,
                 transform: Optional[Callable] = None, pre_transform: Optional[Callable] = None):
        self.name = name
        assert self.name in ['Computers', 'Photos']
        super().__init__(root)
        self.data, self.slices = torch.load(self.processed_paths[0])

    @property
    def raw_dir(self) -> str:
        return osp.join(self.root, self.name, 'raw')
    @property
    def raw_file_names(self) -> List[str]:
        return f'{self.name}.pt'
    @property
    def processed_dir(self) -> str:
        return osp.join(self.root, self.name, 'processed')

    @property
    def processed_file_names(self) -> str:
        return 'data.pt'

    def download(self):
        download_url(self.url.format(self.file_id[self.name]), self.raw_dir, filename=self.raw_file_names)

    def process(self):
        print("Processing the data...")
        raw_data = torch.load(osp.join(self.raw_dir, self.raw_file_names))
        num_hyperedges = torch.unique(raw_data.hyperedge_index[1])
        data = Data(x=raw_data.x, edge_index=raw_data.edge_index, y=raw_data.y.long(),
                    hyperedge_index=raw_data.hyperedge_index, num_hyperedges=len(num_hyperedges))

        torch.save(self.collate([data]), self.processed_paths[0])



import torch
import numpy as np
import math
import torch_geometric.utils as utils


def mask_split(dataset, original_mask=True,
               train_portion=0.6, eval_portion=0.2, test_portion=0.2):
    # re split to the train, eval and test mask to 60:20:20
    # only suppose to work with cora
    masks = []
    for data in dataset:
        if original_mask:
            train_mask = data.train_mask
            val_mask = data.val_mask
            test_mask = data.test_mask
        else:
            num_nodes = data.num_nodes
            indexes = np.arange(num_nodes)
            np.random.shuffle(indexes)
            zeros = torch.zeros((num_nodes))

            train_point = math.ceil(num_nodes * train_portion)
            eval_point = math.ceil(num_nodes * (train_portion + eval_portion))

            train_mask = zeros.clone()
            val_mask = zeros.clone()
            test_mask = zeros.clone()
            train_mask[indexes[:train_point]] = 1
            val_mask[indexes[train_point:eval_point]] = 1
            test_mask[indexes[eval_point:]] = 1
            data.train_mask = train_mask
            data.val_mask = val_mask
            data.test_mask = test_mask

            train_mask = train_mask.type(torch.bool)
            val_mask = val_mask.type(torch.bool)
            test_mask = test_mask.type(torch.bool)

        masks.append((train_mask, val_mask, test_mask))
    return dataset, masks


def random_node_split(data, train_ratio=0.6, val_ratio=0.2, keep_joint_hyperedges=True):
    num_nodes = data.num_nodes
    num_train = int(train_ratio * num_nodes)
    num_val = int(val_ratio * num_nodes)

    train_mask = torch.zeros(num_nodes)
    val_mask = torch.zeros(num_nodes)
    test_mask = torch.zeros(num_nodes)

    perm = torch.randperm(num_nodes)
    train_mask[perm[:num_train]] = 1
    val_mask[perm[num_train:num_train + num_val]] = 1
    test_mask[perm[num_train + num_val:]] = 1

    data.train_mask = train_mask.type(torch.bool)
    data.val_mask = val_mask.type(torch.bool)
    data.test_mask = test_mask.type(torch.bool)

    if keep_joint_hyperedges:
        # If keep_joint_hyperedges is set to True, then a hyperedge is included in a split dataset
        # as long as it contains a node in that split dataset.
        data.train_hyperedge_index = data.hyperedge_index[:, data.train_mask[data.hyperedge_index[0, :]]]
        data.val_hyperedge_index = data.hyperedge_index[:, data.val_mask[data.hyperedge_index[0, :]]]
        data.test_hyperedge_index = data.hyperedge_index[:, data.test_mask[data.hyperedge_index[0, :]]]
        data.num_train_hyperedges = data.train_hyperedge_index[1].unique().numel()
        data.num_val_hyperedges = data.val_hyperedge_index[1].unique().numel()
        data.num_test_hyperedges = data.test_hyperedge_index[1].unique().numel()
    else:
        # If keep_joint_hyperedges is set to False, then a hyperedge is included in a split dataset
        # only if every node in that hyperedge belong to that split dataset. This would drastically
        # reduce the number of total hyperedges in all splits.
        hyperedge_index = data.hyperedge_index
        train_node_in_hyperedge_mask = torch.isin(hyperedge_index[0, :], perm[:num_train])
        val_node_in_hyperedge_mask = torch.isin(hyperedge_index[0, :], perm[num_train:num_train + num_val])
        test_node_in_hyperedge_mask = torch.isin(hyperedge_index[0, :], perm[num_train + num_val:])

        train_hyperedge_to_exclude_mask = hyperedge_index[1, ~train_node_in_hyperedge_mask].unique()
        val_hyperedge_to_exclude_mask = hyperedge_index[1, ~val_node_in_hyperedge_mask].unique()
        test_hyperedge_to_exclude_mask = hyperedge_index[1, ~test_node_in_hyperedge_mask].unique()

        train_hyperedge_mask = ~torch.isin(hyperedge_index[1, :], train_hyperedge_to_exclude_mask)
        val_hyperedge_mask = ~torch.isin(hyperedge_index[1, :], val_hyperedge_to_exclude_mask)
        test_hyperedge_mask = ~torch.isin(hyperedge_index[1, :], test_hyperedge_to_exclude_mask)

        data.train_hyperedge_index = hyperedge_index[:, train_hyperedge_mask]
        data.val_hyperedge_index = hyperedge_index[:, val_hyperedge_mask]
        data.test_hyperedge_index = hyperedge_index[:, test_hyperedge_mask]
        data.num_train_hyperedges = data.num_hyperedges - train_hyperedge_to_exclude_mask.numel()
        data.num_val_hyperedges = data.num_hyperedges - val_hyperedge_to_exclude_mask.numel()
        data.num_test_hyperedges = data.num_hyperedges - test_hyperedge_to_exclude_mask.numel()

    return data

def create_edge_label(datasets,train_ratio=0.6, val_ratio=0.2):
        r"""This is used for edge prediction tasks
        Creates edge labels :obj:`data.edge_label` based on node labels
        :obj:`data.y`.

        Args:
            data (torch_geometric.data.Data): The graph data object.

        :rtype: :class:`torch_geometric.data.Data`
        """
        def edge_label(data):
            # we split the edge_index with all mode
            # get the edge index for train_message_ps = eval_mp,
            # train_supervision + eval_supervision + test_supervision = whole graph
            # test_mp = train_supervision + eval_supervision
            # data.edge_index = utils.to_undirected(data.edge_index)
            train_pos_edge_index, val_pos_edge_index, test_pos_edge_index  = random_edge_split(data,train_ratio, val_ratio)
            data.train_edge_index = data.val_edge_index = train_pos_edge_index
            data.test_edge_index = torch.cat([train_pos_edge_index, val_pos_edge_index], dim=1)
            data.train_label, data.train_edge_label_index = helper(train_pos_edge_index)
            data.val_label, data.val_edge_label_index = helper(val_pos_edge_index)
            data.test_label, data.test_edge_label_index = helper(test_pos_edge_index)
            return data
        def helper(edge_index):
            neg_edge_index = utils.negative_sampling(edge_index)
            # create edge_label
            pos_label = torch.ones(edge_index.size(1), dtype=torch.int64)
            neg_label = torch.zeros(neg_edge_index.size(1), dtype=torch.int64)
            edge_label = torch.cat([pos_label, neg_label], dim=0)
            return edge_label, torch.cat([edge_index, neg_edge_index], dim=1)
        return [edge_label(data) for data in datasets]

def random_edge_split(data, train_ratio=0.6, val_ratio=0.2):
        edge_index = data.edge_index
        num_edges = edge_index.size(1)
        num_train = int(train_ratio * num_edges)
        num_val = int(val_ratio * num_edges)
        perm = torch.randperm(num_edges)
        train_edges = edge_index[:, perm[:num_train]]
        val_edges = edge_index[:, perm[num_train:num_train + num_val]]
        test_edges = edge_index[:, perm[num_train + num_val:]]
        return train_edges, val_edges, test_edges



from .graph_saint import (HypergraphSAINTSampler, HypergraphSAINTNodeSampler, HypergraphSAINTEdgeSampler,
                          HypergraphSAINTRandomWalkSampler)
from .random_sampler import RandomNodeSampler, RandomHyperedgeSampler

__all__ = [
    'HypergraphSAINTSampler',
    'HypergraphSAINTNodeSampler',
    'HypergraphSAINTEdgeSampler',
    'HypergraphSAINTRandomWalkSampler'
    'RandomNodeSampler',
    'RandomHyperedgeSampler',
]



import types
import functools

import torch


use_cuda = torch.cuda.is_available()
print('Using cuda:{}'.format(use_cuda))
torch_cuda = torch.cuda if use_cuda else torch
device = torch.device('cuda' if use_cuda else 'cpu')



import pytorch_lightning as pl
import torch
import numpy as np
from torchmetrics.functional import accuracy
from torchmetrics import MeanAbsoluteError, Accuracy
from torch.optim.lr_scheduler import CosineAnnealingLR


class ModelWrapper(pl.LightningModule):

    def __init__(
            self,
            model,
            dataset_info,
            learning_rate=5e-4,
            epochs=200,
            optimizer=None
            ):
        super().__init__()
        self.model = model
        self.learning_rate = learning_rate
        if dataset_info["is_regression"]:
            self.loss = torch.nn.MSELoss()
            self.train_acc = MeanAbsoluteError()
            self.val_acc = MeanAbsoluteError()
            self.test_acc = MeanAbsoluteError()
        elif dataset_info["is_edge_pred"]:
            self.loss = torch.nn.BCEWithLogitsLoss()
            self.train_acc = Accuracy(task='binary')
            self.val_acc = Accuracy(task='binary')
            self.test_acc = Accuracy(task='binary')
        else:
            self.train_acc = Accuracy(task='multiclass',top_k=1,num_classes=dataset_info["num_classes"])
            self.val_acc = Accuracy(task='multiclass',top_k=1,num_classes=dataset_info["num_classes"])
            self.test_acc = Accuracy(task='multiclass',top_k=1,num_classes=dataset_info["num_classes"])
            self.loss = torch.nn.CrossEntropyLoss()
        self.epochs = epochs
        self.optimizer = optimizer
        self.train_losses = []
        self.val_losses = []
        self.dataset_info = dataset_info

    def forward(self, x, *args, **kwargs):
        return self.model(x, args, kwargs)

    def training_step(self, batch, batch_idx):
        x, y = batch, batch.y
        y = batch.train_label.float() if self.dataset_info["is_edge_pred"] else y
        mask = batch.train_mask
        args = [batch.train_edge_index, batch.train_edge_label_index] if self.dataset_info["is_edge_pred"] else []
        y_hat = self.forward(x,*args)
        loss = self.loss(y_hat,y) if self.dataset_info["is_edge_pred"] else self.loss(y_hat[mask], y[mask])
        #acc = accuracy(y_hat[mask], y[mask],task='multiclass',top_k=1,num_classes=self.dataset_info["num_classes"]) if not self.dataset_info["is_regression"] else self.train_acc(y_hat[mask], y[mask])
        acc = self.train_acc(y_hat[:10], y[:10]) if self.dataset_info["is_edge_pred"] else self.train_acc(y_hat[mask], y[mask])
        # loss
        self.log_dict(
            {"loss": loss},
            on_step=False, on_epoch=True, prog_bar=False, logger=True)
        # acc
        self.log_dict(
            {"acc": acc},
            on_step=False, on_epoch=True, prog_bar=True, logger=True)
        return {"loss": loss, "acc": acc}

    def validation_step(self, batch, batch_idx):
        x, y = batch, batch.y
        y = batch.val_label.float() if self.dataset_info["is_edge_pred"] else y
        mask = batch.val_mask
        args = [batch.val_edge_index, batch.val_edge_label_index] if self.dataset_info["is_edge_pred"] else []
        y_hat = self.forward(x,*args)
        #print(f"y_hat shape: {y_hat.shape}; y shape: {y.shape}")
        loss = self.loss(y_hat,y) if self.dataset_info["is_edge_pred"] else self.loss(y_hat[mask], y[mask])
        #acc = accuracy(y_hat[mask], y[mask],task='multiclass',top_k=1,num_classes=self.dataset_info["num_classes"]) if not self.dataset_info["is_regression"] else self.val_acc(y_hat[mask], y[mask])
        acc = self.train_acc(y_hat, y) if self.dataset_info["is_edge_pred"] else self.train_acc(y_hat[mask], y[mask])
        # val_loss
        self.log_dict(
            {"val_loss": loss},
            on_step=False, on_epoch=True, prog_bar=True, logger=True)
        # val acc
        self.log_dict(
            {"val_acc": acc},
            on_step=False, on_epoch=True, prog_bar=True, logger=True)
        return {"val_loss": loss, "val_acc": acc}

    def test_step(self, batch, batch_idx):
        x, y = batch, batch.y
        y = batch.test_label.float() if self.dataset_info["is_edge_pred"] else y
        mask = batch.test_mask
        args = [batch.test_edge_index, batch.test_edge_label_index] if self.dataset_info["is_edge_pred"] else []
        y_hat = self.forward(x,*args)
        loss = self.loss(y_hat,y) if self.dataset_info["is_edge_pred"] else self.loss(y_hat[mask], y[mask])
        #acc = accuracy(y_hat[mask], y[mask],task='multiclass',top_k=1,num_classes=self.dataset_info["num_classes"]) if not self.dataset_info["is_regression"] else self.test_acc(y_hat[mask], y[mask])
        acc = self.train_acc(y_hat, y) if self.dataset_info["is_edge_pred"] else self.train_acc(y_hat[mask], y[mask])
        # val_loss
        self.log_dict(
            {"test_loss": loss},
            on_step=False, on_epoch=True, prog_bar=False, logger=True)
        # val acc
        self.log_dict(
            {"test_acc": acc},
            on_step=False, on_epoch=True, prog_bar=True, logger=True)
        return {"test_loss": loss, "test_acc": acc}

    def configure_optimizers(self):
        if self.optimizer == 'adam':
            opt = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
            scheduler = CosineAnnealingLR(opt, T_max=self.epochs, eta_min=1e-6)
        elif self.optimizer in ['sgd_warmup', 'sgd']:
            opt = torch.optim.SGD(
                self.parameters(),
                lr=self.learning_rate,
                momentum=0.9,
                weight_decay=0.0005,
                nesterov=True)
            if self.optimizer == 'sgd':
                scheduler = CosineAnnealingLR(
                    opt, T_max=self.epochs, eta_min=0.0)
            elif self.optimizer == 'sgd_warmup':
                # mostly only useful for large scale image datasets
                from warmup_scheduler import GradualWarmupScheduler
                scheduler = CosineAnnealingLR(
                    opt, T_max=self.epochs, eta_min=0.0)
                scheduler = GradualWarmupScheduler(
                    opt,
                    multiplier=2,
                    total_epoch=5,
                    after_scheduler=scheduler)
        return {
            "optimizer": opt,
            "lr_scheduler":  scheduler}



import os.path as osp
from typing import List, Callable, Optional
import torch
from torch_geometric.data import InMemoryDataset, download_url, Data
import torch_geometric.utils as utils


class Grand(InMemoryDataset):
    r"""The Grand dataset
    Args:
        root (str): Root directory where the dataset should be saved.
        num_graphs (int): The number of graphs to be loaded.
     **STATS:**

    .. list-table::
        :widths: 15 15 15 15 15
        :header-rows: 1

        * - Name
          - #nodes
          - #edges
          - #hyperedges
          - #features
          - #classes
        * - Leukemia
          - 4,651
          - 6,362
          - 7,812
          - 4608
          - 3
        * - Brain
          - 6,196
          - 6,245
          - 11,878
          - 4608
          - 3
        * - Kidney_renal_papillary_cell_carcinoma
          - 4,319
          - 5,599
          - 7,369
          - 4608
          - 3
        * - Lung
          - 6,119
          - 6,160
          - 11,760
          - 4608
          - 3
        * - Breast
          - 5,921
          - 5,910
          - 11,400
          - 4608
          - 3
        * - Artery_Coronary
          - 5,755
          - 5,722
          - 11,222
          - 4608
          - 3
        * - Artery_Aorta
          - 5,848
          - 5,823
          - 11,368
          - 4608
          - 3
        * - Lung_cancer
          - 4,896
          - 6,995
          - 8,179
          - 4608
          - 3
        * - Stomach_cancer
          - 4,518
          - 6,051
          - 7,611
          - 4608
          - 3
        * - Stomach
          - 5,745
          - 5,694
          - 11,201
          - 4608
          - 3
        """
    url = "https://drive.google.com/uc?export=download&confirm=no_antivirus&id={}"
    file_id = {
        'Leukemia': '1ztRJcNXG4O6OY106JFyZ8MF1EyV55n1_',
        'Brain': '1kBxVOj_H0FYRc159HypgvLw5zwsp-9Pv',
        'Kidney_renal_papillary_cell_carcinoma': '1Hj5zZFOon3YW3MTR0vzDrfS0IqgkwcNH',
        'Lung': '14g1wbCRbepZaniRdsCD-TP0p6SPTmvhS',
        'Breast': '1cinfmlomxkenscWZwTCaXy8pyGdx78gs',
        'Artery_Coronary': '1LyvWdPGXMelvVmh2g_RS3iRnUgBIJNf-',
        'Artery_Aorta': '1XjTRe4DJ2RU8SlqHkgUa6NTOdMLkGsd_',
        'Lung_cancer': '1QLRHsyhssELyC4eJ9sH75E3QJ5lHUNrA',
        'Stomach_cancer':'1TF0rg3agGfI9ScZijcQgH6FhkZ6O6EsB',
        'Stomach':'1MWc_wINEogQHX5u6xvvOWyohtd_HIwoj',
    }
    def __init__(self,  root: str, name: str,
                 transform: Optional[Callable] = None, pre_transform: Optional[Callable] = None):
        self.name = name
        assert self.name in ['Artery_Aorta','Breast','Artery_Coronary','Stomach_cancer',
                             'Stomach', 'Brain','Lung',
                             'Kidney_renal_papillary_cell_carcinoma',
                             'Lung_cancer','Leukemia']
        super().__init__(root)
        self.data, self.slices = torch.load(self.processed_paths[0])

    @property
    def raw_dir(self) -> str:
        return osp.join(self.root, self.name, 'raw')
    @property
    def raw_file_names(self) -> List[str]:
        return f'{self.name}.pt'
    @property
    def processed_dir(self) -> str:
        return osp.join(self.root, self.name, 'processed')

    @property
    def processed_file_names(self) -> str:
        return 'data.pt'

    def download(self):
        download_url(self.url.format(self.file_id[self.name]), self.raw_dir, filename=self.raw_file_names)

    def process(self):
        print("Processing the data...")
        raw_data = torch.load(osp.join(self.raw_dir, self.raw_file_names))
        # torch.arange(30171).view(-1,1).float()
        data = Data(x=raw_data.x, edge_index=raw_data.edge_index, y=raw_data.y,
                    hyperedge_index=raw_data.hyperedge_index, num_hyperedges=raw_data.num_hyperedges)

        torch.save(self.collate([data]), self.processed_paths[0])

    def to_undirected(self, edge_index):
        # Given an edge_index tensor of shape [2, E], we first create a new tensor
        # that includes both the original edges and their reverse.
        undirected_edge_index = torch.cat([edge_index, torch.flip(edge_index, [0])], dim=1)

        # In the new tensor, there may be duplicate edges, so we call unique to remove them.
        undirected_edge_index = torch.unique(undirected_edge_index, dim=1)

        return undirected_edge_index



from typing import Optional

import torch
from torch_geometric.loader import GraphSAINTSampler


class HypergraphSAINTSampler(GraphSAINTSampler):
    # def _collate(self, data_list):  # torch-geometric>=2.3.0
    def __collate__(self, data_list):  # torch-geometric==2.2.0
        assert len(data_list) == 1
        node_idx, adj = data_list[0]

        data = self.data.__class__()
        data.num_nodes = node_idx.size(0)
        row, col, edge_idx = adj.coo()
        data.edge_index = torch.stack([row, col], dim=0)
        hyperedge_mask = torch.isin(self.data.hyperedge_index[0], node_idx)
        data.hyperedge_index = self.data.hyperedge_index[:, hyperedge_mask]
        _, inverse_indices = torch.unique(data.hyperedge_index[1], return_inverse=True)
        data.hyperedge_index[1] = inverse_indices
        data.num_hyperedges = data.hyperedge_index[1, -1] + 1

        # Remap node indices in hyperedge_index to the range [0, sample_size - 1]
        node_map = {node: i for i, node in enumerate(node_idx.tolist())}
        data.hyperedge_index[0].apply_(lambda node: node_map[node])

        for key, item in self.data:
            if key in ['edge_index', 'hyperedge_index', 'num_nodes', 'num_hyperedges']:
                continue
            if isinstance(item, torch.Tensor) and item.size(0) == self.N:
                data[key] = item[node_idx]
            elif isinstance(item, torch.Tensor) and item.size(0) == self.E:
                data[key] = item[edge_idx]
            else:
                data[key] = item

        if self.sample_coverage > 0:
            data.node_norm = self.node_norm[node_idx]
            data.edge_norm = self.edge_norm[edge_idx]

        return data


class HypergraphSAINTNodeSampler(HypergraphSAINTSampler):
    # def _sample_nodes(self, batch_size):  # torch-geometric>=2.3.0
    def __sample_nodes__(self, batch_size):  # torch-geometric==2.2.0
        edge_sample = torch.randint(0, self.E, (batch_size, self.batch_size), dtype=torch.long)

        return self.adj.storage.row()[edge_sample]


class HypergraphSAINTEdgeSampler(HypergraphSAINTSampler):
    # def _sample_nodes(self, batch_size):  # torch-geometric>=2.3.0
    def __sample_nodes__(self, batch_size):  # torch-geometric==2.2.0
        row, col, _ = self.adj.coo()

        deg_in = 1. / self.adj.storage.colcount()
        deg_out = 1. / self.adj.storage.rowcount()
        prob = (1. / deg_in[row]) + (1. / deg_out[col])

        # Parallel multinomial sampling (without replacement)
        # https://github.com/pytorch/pytorch/issues/11931#issuecomment-625882503
        rand = torch.rand(batch_size, self.E).log() / (prob + 1e-10)
        edge_sample = rand.topk(self.batch_size, dim=-1).indices

        source_node_sample = col[edge_sample]
        target_node_sample = row[edge_sample]

        return torch.cat([source_node_sample, target_node_sample], -1)


class HypergraphSAINTRandomWalkSampler(HypergraphSAINTSampler):
    def __init__(self, data, batch_size: int, walk_length: int,
                 num_steps: int = 1, sample_coverage: int = 0,
                 save_dir: Optional[str] = None, log: bool = True, **kwargs):
        self.walk_length = walk_length
        super().__init__(data, batch_size, num_steps, sample_coverage,
                         save_dir, log, **kwargs)

    @property
    # def _filename(self):  # torch-geometric>=2.3.0
    def __filename__(self):  # torch-geometric==2.2.0
        return (f'{self.__class__.__name__.lower()}_{self.walk_length}_'
                f'{self.sample_coverage}.pt')

    # def _sample_nodes(self, batch_size):  # torch-geometric>=2.3.0
    def __sample_nodes__(self, batch_size):  # torch-geometric==2.2.0
        start = torch.randint(0, self.N, (batch_size,), dtype=torch.long)
        node_idx = self.adj.random_walk(start.flatten(), self.walk_length)
        return node_idx.view(-1)



import yaml
import torch
import torch.nn.functional as F
import torch.nn as nn
from torch_geometric.nn import GCNConv, HypergraphConv, GAT
from .attention import Attention


class LPGCNHyperGCN(torch.nn.Module):
    def __init__(
            self, info, *args, **kwargs):
        super().__init__()
        dim = 64
        self.hyper1 = HypergraphConv(info["num_node_features"], dim)
        self.conv1 = GCNConv(info["num_node_features"], dim)
        self.is_regression = info["is_regression"]
        if info["is_regression"]:
            self.hyper2 = HypergraphConv(dim, dim)
            self.conv2 = GCNConv(dim, dim)
            self.lp = nn.Linear(2*dim,dim) # TODO: cross attention between q = hyper2(x) and k = conv2(x)
            self.head = nn.Linear(dim, 1)
        else:
            self.conv2 = GCNConv(dim, info["num_classes"])
            self.hyper2 = HypergraphConv(dim, info["num_classes"])
            self.lp = nn.Linear(2*info["num_classes"],info["num_classes"])
    def forward(self, data, *args, **kargs):
        x, edge_index,hyperedge_index = data.x, data.edge_index, data.hyperedge_index
        x_gcn = F.relu(self.conv1(x, edge_index))
        x_hyper = F.relu(self.hyper1(x, hyperedge_index))
        x_gcn = F.dropout(x_gcn, training=self.training)
        x_hyper = F.dropout(x_hyper, training=self.training)
        x_gcn = self.conv2(x_gcn, edge_index)
        x_hyper = self.hyper2(x_hyper, hyperedge_index)
        x = self.lp(torch.cat([x_gcn.unsqueeze(0),x_hyper.unsqueeze(0)],dim=2)).squeeze(0)

        if self.is_regression:
            x = self.head(x).squeeze()
        else:
            x = F.log_softmax(x, dim=1)
        return x
    

class LPGATHyperGCN(torch.nn.Module):
    def __init__(
            self, info, *args, **kwargs):
        super().__init__()
        dim = 64
        self.hyper1 = HypergraphConv(info["num_node_features"], dim)
        self.conv1 = GAT(info["num_node_features"], dim,num_layers=1)
        self.is_regression = info["is_regression"]
        if info["is_regression"]:
            self.hyper2 = HypergraphConv(dim, dim)
            self.conv2 =GAT(dim, dim,num_layers=1)
            self.lp = nn.Linear(2*dim,dim) # TODO: cross attention between q = hyper2(x) and k = conv2(x)
            self.head = nn.Linear(dim, 1)
        else:
            self.conv2 = GCNConv(dim, info["num_classes"])
            self.hyper2 = GAT(dim, info["num_classes"],num_layers=1)
            self.lp = nn.Linear(2*info["num_classes"],info["num_classes"])
    def forward(self, data, *args, **kargs):
        x, edge_index,hyperedge_index = data.x, data.edge_index, data.hyperedge_index
        x_gcn = F.relu(self.conv1(x, edge_index))
        x_hyper = F.relu(self.hyper1(x, hyperedge_index))
        x_gcn = F.dropout(x_gcn, training=self.training)
        x_hyper = F.dropout(x_hyper, training=self.training)
        x_gcn = self.conv2(x_gcn, edge_index)
        x_hyper = self.hyper2(x_hyper, hyperedge_index)
        x = self.lp(torch.cat([x_gcn.unsqueeze(0),x_hyper.unsqueeze(0)],dim=2)).squeeze(0)

        if self.is_regression:
            x = self.head(x).squeeze()
        else:
            x = F.log_softmax(x, dim=1)
        return x


class LPGGATGCN(torch.nn.Module):
    def __init__(
            self, info, *args, **kwargs):
        super().__init__()
        dim = 64
        self.hyper1 = GAT(info["num_node_features"], dim,num_layers=1)
        self.conv1 = GCNConv(info["num_node_features"], dim)
        self.is_regression = info["is_regression"]
        if info["is_regression"]:
            self.hyper2 = GAT(dim, dim,num_layers=1 )
            self.conv2 = GCNConv(dim, dim)
            self.lp = nn.Linear(2*dim,dim) # TODO: cross attention between q = hyper2(x) and k = conv2(x)
            self.head = nn.Linear(dim, 1)
        else:
            self.conv2 = GCNConv(dim, info["num_classes"])
            self.hyper2 = GAT(dim, info["num_classes"],num_layers=1 )
            self.lp = nn.Linear(2*info["num_classes"],info["num_classes"])
    def forward(self, data, *args, **kargs):
        x, edge_index,hyperedge_index = data.x, data.edge_index, data.hyperedge_index
        x_gcn = F.relu(self.conv1(x, edge_index))
        x_hyper = F.relu(self.hyper1(x, edge_index))
        x_gcn = F.dropout(x_gcn, training=self.training)
        x_hyper = F.dropout(x_hyper, training=self.training)
        x_gcn = self.conv2(x_gcn, edge_index)
        x_hyper = self.hyper2(x_hyper, edge_index)
        x = self.lp(torch.cat([x_gcn.unsqueeze(0),x_hyper.unsqueeze(0)],dim=2)).squeeze(0)

        if self.is_regression:
            x = self.head(x).squeeze()
        else:
            x = F.log_softmax(x, dim=1)
        return x
    

class LPGGATGAT(torch.nn.Module):
    def __init__(
            self, info, *args, **kwargs):
        super().__init__()
        dim = 64
        self.hyper1 = GAT(info["num_node_features"], dim,num_layers=1)
        self.conv1 =  GAT(info["num_node_features"], dim,num_layers=1)
        self.is_regression = info["is_regression"]
        if info["is_regression"]:
            self.hyper2 = GAT(dim, dim,num_layers=1 )
            self.conv2 =  GAT(dim, dim,num_layers=1 )
            self.lp = nn.Linear(2*dim,dim) # TODO: cross attention between q = hyper2(x) and k = conv2(x)
            self.head = nn.Linear(dim, 1)
        else:
            self.conv2 = GAT(dim, info["num_classes"],num_layers=1 )
            self.hyper2 = GAT(dim, info["num_classes"],num_layers=1 )
            self.lp = nn.Linear(2*info["num_classes"],info["num_classes"])
    def forward(self, data, *args, **kargs):
        x, edge_index,hyperedge_index = data.x, data.edge_index, data.hyperedge_index
        x_gcn = F.relu(self.conv1(x, edge_index))
        x_hyper = F.relu(self.hyper1(x, edge_index))
        x_gcn = F.dropout(x_gcn, training=self.training)
        x_hyper = F.dropout(x_hyper, training=self.training)
        x_gcn = self.conv2(x_gcn, edge_index)
        x_hyper = self.hyper2(x_hyper, edge_index)
        x = self.lp(torch.cat([x_gcn.unsqueeze(0),x_hyper.unsqueeze(0)],dim=2)).squeeze(0)

        if self.is_regression:
            x = self.head(x).squeeze()
        else:
            x = F.log_softmax(x, dim=1)
        return x
    

class LPGCNGCN(torch.nn.Module):
    def __init__(
            self, info, *args, **kwargs):
        super().__init__()
        dim = 64
        self.hyper1 = GCNConv(info["num_node_features"], dim)
        self.conv1 = GCNConv(info["num_node_features"], dim)
        self.is_regression = info["is_regression"]
        if info["is_regression"]:
            self.hyper2 = GCNConv(dim, dim)
            self.conv2 =  GCNConv(dim, dim)
            self.lp = nn.Linear(2*dim,dim) # TODO: cross attention between q = hyper2(x) and k = conv2(x)
            self.head = nn.Linear(dim, 1)
        else:
            self.hyper2 = GCNConv(dim, info["num_classes"])
            self.conv2 = GCNConv(dim, info["num_classes"])
            self.lp = nn.Linear(2*info["num_classes"],info["num_classes"])
    def forward(self, data, *args, **kargs):
        x, edge_index,hyperedge_index = data.x, data.edge_index, data.hyperedge_index
        x_gcn = F.relu(self.conv1(x, edge_index))
        x_hyper = F.relu(self.hyper1(x, edge_index))
        x_gcn = F.dropout(x_gcn, training=self.training)
        x_hyper = F.dropout(x_hyper, training=self.training)
        x_gcn = self.conv2(x_gcn, edge_index)
        x_hyper = self.hyper2(x_hyper, edge_index)
        x = self.lp(torch.cat([x_gcn.unsqueeze(0),x_hyper.unsqueeze(0)],dim=2)).squeeze(0)

        if self.is_regression:
            x = self.head(x).squeeze()
        else:
            x = F.log_softmax(x, dim=1)
        return x
    

class LPHYPERHYPER(torch.nn.Module):
    def __init__(
            self, info, *args, **kwargs):
        super().__init__()
        dim = 64
        self.hyper1 = HypergraphConv(info["num_node_features"], dim)
        self.conv1 = HypergraphConv(info["num_node_features"], dim)
        self.is_regression = info["is_regression"]
        if info["is_regression"]:
            self.hyper2 = HypergraphConv(dim, dim)
            self.conv2 =HypergraphConv(dim, dim)
            self.lp = nn.Linear(2*dim,dim) # TODO: cross attention between q = hyper2(x) and k = conv2(x)
            self.head = nn.Linear(dim, 1)
        else:
            self.conv2 = HypergraphConv(dim, info["num_classes"])
            self.hyper2 = HypergraphConv(dim, info["num_classes"])
            self.lp = nn.Linear(2*info["num_classes"],info["num_classes"])
    def forward(self, data, *args, **kargs):
        x, edge_index,hyperedge_index = data.x, data.edge_index, data.hyperedge_index
        x_gcn = F.relu(self.conv1(x, hyperedge_index))
        x_hyper = F.relu(self.hyper1(x, hyperedge_index))
        x_gcn = F.dropout(x_gcn, training=self.training)
        x_hyper = F.dropout(x_hyper, training=self.training)
        x_gcn = self.conv2(x_gcn, hyperedge_index)
        x_hyper = self.hyper2(x_hyper, hyperedge_index)
        x = self.lp(torch.cat([x_gcn.unsqueeze(0),x_hyper.unsqueeze(0)],dim=2)).squeeze(0)

        if self.is_regression:
            x = self.head(x).squeeze()
        else:
            x = F.log_softmax(x, dim=1)
        return x



import yaml
import torch
import torch.nn.functional as F
import torch.nn as nn
from torch_geometric.nn import HypergraphConv

"""
for baseline method:
# self.hyperGCN
# x2 = hyperGCN(x,hyperedge_index)
# x,x2
# potential 2 impelmentations
"""
class HyperGCN(torch.nn.Module):
    def __init__(
            self, info, *args, **kwargs):
        super().__init__()
        dim = 32
        self.conv1 = HypergraphConv(info["num_node_features"], dim)
        self.is_regression = info["is_regression"]
        self.is_edge_pred = info["is_edge_pred"]
        if info["is_regression"]:
            self.conv2 = HypergraphConv(dim, dim, )
            self.head = nn.Linear(dim, 1)
        elif info["is_edge_pred"]:
            self.conv2 = HypergraphConv(dim, dim, )
        else:
            self.conv2 = HypergraphConv(dim, info["num_classes"])

    def forward(self, data, *args, **kargs):
        x, edge_index = data.x, data.hyperedge_index
        if self.is_edge_pred:
            edge_index = args[0][0] # the message passing edge index
        x = F.relu(self.conv1(x, edge_index))
        x = F.dropout(x, training=self.training)
        x = self.conv2(x, edge_index)


        if self.is_regression:
            x = self.head(x).squeeze()
        elif self.is_edge_pred:
            # args[1] is the edge_label_index
            edge_label_index = args[0][1]
            x = x[edge_label_index[0]]*x[edge_label_index[1]]
            x = torch.sum(x,-1)
        else:
            x = F.log_softmax(x, dim=1)
        return x


class HyperGAT(torch.nn.Module):
    def __init__(
            self, info, mixture_cls=None, load_config=None, *args, **kwargs):
        super().__init__()
        dim = 32
        heads = info["num_node_features"]//dim
        self.is_regression = info["is_regression"]
        self.is_edge_pred = info["is_edge_pred"]
        if self.is_regression:
            self.conv2 = HypergraphConv(dim*heads, dim, use_attention=True)
            self.head = nn.Linear(dim, 1)
        elif info["is_edge_pred"]:
            self.conv2 = HypergraphConv(dim*heads, dim, use_attention=True)
        else:
            self.conv2 = HypergraphConv(dim*heads, info["num_classes"], use_attention=True)

        self.conv1 = HypergraphConv(
            info["num_node_features"], dim, use_attention=True,heads=heads)

    def forward(self, data, *args, **kargs):
        x, edge_index = data.x, data.hyperedge_index
        if self.is_edge_pred:
            edge_index = args[0][0] # the message passing edge index
        x = F.relu(self.conv1(x, edge_index,
                              hyperedge_attr=self.hyperedge_representation(x, edge_index)))
        x = F.dropout(x, p=0.5, training=self.training)

        x = self.conv2(x, edge_index, hyperedge_attr=self.hyperedge_representation(x, edge_index))
        if self.is_regression:
            x = self.head(x).squeeze()
        elif self.is_edge_pred:
            # args[1] is the edge_label_index
            edge_label_index = args[0][1]
            x = x[edge_label_index[0]]*x[edge_label_index[1]]
            x = torch.sum(x,-1)
        else:
            x = F.log_softmax(x, dim=1)
        return x

    def hyperedge_representation(self, x, hyperedge_index):
        # Assuming x is a NumPy array, if not, convert it to a NumPy array
        # Initialize a placeholder array with the length of the maximum hyperedge index + 1
        max_hyperedge_index = torch.max(hyperedge_index[1])
        hyperedge_representations = torch.zeros((max_hyperedge_index + 1, x.shape[1]),device=x.device)
         # Use torch_scatter's scatter_add to accumulate the node values for each hyperedge
        from torch_scatter import scatter_add
        hyperedge_representations = scatter_add(src=x[hyperedge_index[0]], index=hyperedge_index[1], dim=0, out=hyperedge_representations)
        """
        # Iterate through hyperedge_index
        for i in range(hyperedge_index.shape[1]):
            source, target = hyperedge_index[:, i]

            # Sum the node values for each hyperedge
            hyperedge_representations[target] += x[source]
        """
        return hyperedge_representations



import os
import torch
import pytorch_lightning as pl

from .plt_wrapper import ModelWrapper


def get_checkpoint_file(checkpoint_dir):
    for file in os.listdir(checkpoint_dir):
        if file.endswith(".ckpt"):
            return file


def plt_model_load(model, checkpoint):
    state_dict = torch.load(checkpoint)['state_dict']
    model.load_state_dict(state_dict)
    return model


def test(model, test_loader, plt_trainer_args, load_path, dataset_info):
    plt_model = ModelWrapper(model, dataset_info)
    print(load_path)
    if load_path is not None:
        if load_path.endswith(".ckpt"):
            checkpoint = load_path
        else:
            if load_path.endswith("/"):
                checkpoint = load_path + "best.ckpt"
            else:
                raise ValueError(
                    "if it is a directory, if must end with /; if it is a file, it must end with .ckpt")
        plt_model = plt_model_load(plt_model, checkpoint)
        print(f"Loaded model from {checkpoint}")
    #randomize_weights(plt_model)
    trainer = pl.Trainer(**plt_trainer_args)
    trainer.test(plt_model, test_loader)

def randomize_weights(module, a=-1, b=1):
    """
    Randomize weights of a given PyTorch module using uniform distribution.

    Args:
        module (nn.Module): The PyTorch module whose weights will be randomized.
        a (float, optional): Lower bound of the uniform distribution. Default is -1.
        b (float, optional): Upper bound of the uniform distribution. Default is 1.
    """
    for name, param in module.named_parameters():
        if param.requires_grad and 'weight' in name:
            print(name)
            torch.nn.init.uniform_(param.data, a=a, b=b)
        elif param.requires_grad and 'bias' in name:
            print(name)
            torch.nn.init.zeros_(param.data)



from setuptools import setup, find_packages

setup(
    name="hybrid-graph",
    version="0.4",
    packages=find_packages(),
    package_data={'hg': ['datasets/dataset_info.yaml']},
    install_requires=[
        'torch',
        'torch_scatter',
        'torch_sparse',
        'torch_geometric==2.2.0',
        'ipdb==0.13.13',
        'ipython==8.13.2',
        'matplotlib==3.7.1',
        'networkx==3.1',
        'numpy==1.24.2',
        'pytorch_lightning==1.9.3',
        'PyYAML==6.0',
        'scikit_learn==1.2.1',
        'seaborn==0.12.2',
        'torchmetrics==0.11.1'],
    entry_points={
        'console_scripts': [
            'hybrid-graph = hg.hybrid_graph.cli:main',
        ],
    },
    author="Xiangyu Zhao;Zehui Li",
    author_email="zehui.li22@imperial.ac.uk, x.zhao22@imperial.ac.uk",
    description="A python package for accessing and Hybrid-graph Datasets and train-eval framework for GNNs",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Zehui127/hybrid-graph-benchmark/",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import torch
import torch_geometric
from torch_geometric.data import Data
from torch_geometric.nn import knn_graph
from torch_geometric.utils import degree


class GraphFormatter(object):
    """
    A class to create a hypergraph from a graph using a threshold value.
    based on the distance between the node embeddings.
    """

    def __init__(self, k=None, percentile=80):
        self.k = k
        self.percentile = percentile

    def __call__(self, data):
        self.k = self.k if self.k is not None else degree(data.edge_index[0]).float().mean().item()
        self.k = int(self.k)
        x, y = data.x, data.y
        if hasattr(data, "emb"):
            hyper_x = data.emb
        else:
            hyper_x = x
        hyper_edge_index = get_hyper_edge_index(hyper_x, self.k, self.percentile)

        new_data = Data(
            x=x,
            edge_index=data.edge_index,
            y=y,
            hyperedge_index=hyper_edge_index,
            num_hyperedges=hyper_edge_index.shape[1],
        )
        if hasattr(data, "train_mask"):
            new_data.train_mask = data.train_mask
        if hasattr(data, "val_mask"):
            new_data.val_mask = data.val_mask
        if hasattr(data, "test_mask"):
            new_data.test_mask = data.test_mask
        return new_data


def compute_pairwise_distances(x, k, loop=True):
    """
    Compute pairwise distances between all the nodes in the graph.

    Args:
        x (torch.Tensor): Node features of shape (num_nodes, num_features).
        k (int): The number of nearest neighbors to consider.
        loop (bool): Whether to include self-loops in the output graph.

    Returns:
        edge_index (torch.Tensor): The edge indices of the k-nearest neighbor graph of shape (2, num_edges).
    """
    edge_index = knn_graph(x, k=k, batch=None, loop=loop)
    return edge_index


def create_point_clouds(x, k, threshold=None):
    """
    Create point clouds using a threshold value.

    Args:
        x (torch.Tensor): Node features of shape (num_nodes, num_features).
        k (int): The number of nearest neighbors to consider.
        threshold (float): The threshold value for edge distances.

    Returns:
        point_clouds (list): A list of point clouds as torch.Tensor.
    """
    edge_index = compute_pairwise_distances(x, k)

    if threshold is not None:
        edge_distances = torch.norm(x[edge_index[0]] - x[edge_index[1]], dim=1)
        mask = edge_distances < threshold
        thresholded_edge_index = edge_index[:, mask]
    else:
        thresholded_edge_index = edge_index

    return thresholded_edge_index, edge_index

def find_threshold(x, k, percentile):
    """
    Find the threshold value based on a certain percentile of the distance distribution.

    Args:
        x (torch.Tensor): Node features of shape (num_nodes, num_features).
        k (int): The number of nearest neighbors to consider.
        percentile (float): The percentile value for the threshold.

    Returns:
        threshold (float): The threshold value.
    """
    edge_index = compute_pairwise_distances(x, k)
    edge_distances = torch.norm(x[edge_index[0]] - x[edge_index[1]], dim=1)
    q = percentile / 100
    threshold = torch.quantile(edge_distances, q)
    return threshold


def get_hyper_edge_index(x, k, percentile):
    threshold = find_threshold(x, k, percentile)
    edge_index, _ = create_point_clouds(x, k, threshold)
    return edge_index
